from enum import Enum


class ModelNumber(Enum):

    NONE = 0x00000000

    #                    AAAABBCC, AAAA(Project Number), BB(Device Type), CC(Revision)
    DRONE_3_DRONE_P1 = 0x00031001       # Lightrone / GD65 / HW2181
    DRONE_3_DRONE_P2 = 0x00031002       # Soccer Drone / HW2181
    DRONE_3_DRONE_P3 = 0x00031003       # GD240 / HW2181
    DRONE_3_DRONE_P4 = 0x00031004       # GD50N / HW2181
    DRONE_3_DRONE_P5 = 0x00031005       # GD30 / HW2181
    DRONE_3_DRONE_P6 = 0x00031006       # Soccer Drone 2 / HW2181
    DRONE_3_DRONE_P7 = 0x00031007       # SKYKICKV2 / HW2181 
    DRONE_3_DRONE_P8 = 0x00031008       # GD65 / HW2181
    DRONE_3_DRONE_P9 = 0x00031009       # GD65 / HW2181
    DRONE_3_DRONE_P10 = 0x0003100A      # Battle Drone / SPI / HW2181

    DRONE_3_CONTROLLER_P1 = 0x00032001  # (obsolete)
    DRONE_3_CONTROLLER_P2 = 0x00032002  # HW2000
    DRONE_3_CONTROLLER_P3 = 0x00032003  # HW2000B
    DRONE_3_CONTROLLER_P4 = 0x00032004
    DRONE_3_CONTROLLER_P5 = 0x00032005

    DRONE_3_LINK_P0 = 0x00033000

    DRONE_3_ANALYZER_P0 = 0x00034100    # for soccer drone

    DRONE_3_TESTER_P4 = 0x0003A004      # (obsolete)
    DRONE_3_TESTER_P6 = 0x0003A006      # Battle Drone Tester

    DRONE_4_DRONE_P4 = 0x00041004       # (obsolete)
    DRONE_4_DRONE_P5 = 0x00041005       # HW2000, 2m range sensor
    DRONE_4_DRONE_P6 = 0x00041006       # HW2000B, 4m range sensor
    DRONE_4_DRONE_P7 = 0x00041007       # HW2000B, 4m range sensor, BLDC Motor

    DRONE_4_CONTROLLER_P1 = 0x00042001  # (obsolete)
    DRONE_4_CONTROLLER_P2 = 0x00042002  # HW2000
    DRONE_4_CONTROLLER_P3 = 0x00042003  # HW2000B
    DRONE_4_CONTROLLER_P4 = 0x00042004  # HW2000B + PA
    DRONE_4_CONTROLLER_P5 = 0x00042005  # HW2000B + PA

    DRONE_4_LINK_P0 = 0x00043000

    DRONE_4_TESTER_P4 = 0x0004A004      # (obsolete)
    DRONE_4_TESTER_P6 = 0x0004A006
    DRONE_4_TESTER_P7 = 0x0004A007

    DRONE_4_MONITOR_P4 = 0x0004A104     # (obsolete)

    DRONE_7_DRONE_P1 = 0x00071001       # (obsolete)
    DRONE_7_DRONE_P2 = 0x00071002       # Coding Car

    DRONE_7_BLE_CLIENT_P0 = 0x00073200  # Coding Car Link
    DRONE_7_BLE_CLIENT_P5 = 0x00073205  # Coding Car Tester BLE

    DRONE_7_BLE_SERVER_P2 = 0x00073302  # Coding Car Ble Module

    DRONE_7_TESTER_P4 = 0x0003A004      # (obsolete)
    DRONE_7_TESTER_P5 = 0x0003A005      # (obsolete)
    DRONE_7_TESTER_P6 = 0x0003A006

    DRONE_7_MONITOR_P4 = 0x0003A104     # (obsolete)
    DRONE_7_MONITOR_P5 = 0x0003A105

    DRONE_8_DRONE_P0 = 0x00081000       # (obsolete)
    DRONE_8_DRONE_P1 = 0x00081001       # Coding Drone

    DRONE_8_TESTER_P4 = 0x0008A004      # (obsolete)
    DRONE_8_TESTER_P6 = 0x0008A006

    DRONE_8_MONITOR_P6 = 0x0008A106

    DRONE_9_DRONE_P0 = 0x00091000
    DRONE_9_DRONE_P1 = 0x00091001
    DRONE_9_DRONE_P2 = 0x00091002
    DRONE_9_DRONE_P3 = 0x00091003
    DRONE_9_DRONE_P4 = 0x00091004
    DRONE_9_DRONE_P5 = 0x00091005
    DRONE_9_DRONE_P6 = 0x00091006

    DRONE_9_TESTER_P6 = 0x0009A006
    DRONE_9_TESTER_P7 = 0x0009A007


class DeviceType(Enum):

    NONE = 0x00

    DRONE = 0x10        # 드론(Server)

    CONTROLLER = 0x20   # 조종기(Client)

    LINK = 0x30         # 링크 모듈(Client)
    LINK_SERVER = 0x31  # 링크 모듈(Server, 링크 모듈이 서버로 동작하는 경우에만 통신 타입을 잠시 바꿈)
    BLE_CLIENT = 0x32   # BLE 클라이언트
    BLE_SERVER = 0x33   # BLE 서버

    RANGE = 0x40        # 거리 센서 모듈
    ANALYZER = 0x41     # 분석 모듈

    BASE = 0x70         # 베이스

    BYSCRATCH = 0x80    # 바이스크래치
    SCRATCH = 0x81      # 스크래치
    ENTRY = 0x82        # 네이버 엔트리

    TESTER = 0xA0       # 테스터
    MONITOR = 0xA1      # 모니터
    UPDATER = 0xA2      # 펌웨어 업데이트 도구
    ENCRYPTER = 0xA3    # 암호화 도구

    # 바로 인접한 장치까지만 전달(받은 장치는 자기 자신에게 보낸 것처럼 처리하고 타 장치에 전달하지 않음)
    WHISPERING = 0xFE
    BROADCASTING = 0xFF


class ModeSystem(Enum):

    NONE = 0x00

    BOOT = 0x10
    START = 0x11
    RUNNING = 0x12
    READY_TO_RESET = 0x13

    ERROR = 0xA0

    END_OF_TYPE = 0xA1


class ModeControlFlight(Enum):

    NONE = 0x00

    ATTITUDE = 0x10     # 자세 - X, Y는 각도(deg)로 입력받음, Z, Yaw는 속도(m/s)로 입력 받음
    POSITION = 0x11     # 위치 - X, Y, Z, Yaw는 속도(m/s)로 입력 받음
    MANUAL = 0x12       # 고도를 수동으로 조종함
    RATE = 0x13         # Rate - X, Y는 각속도(deg/s)로 입력받음, Z, Yaw는 속도(m/s)로 입력 받음
    FUNCTION = 0x14     # 기능

    END_OF_TYPE = 0x15


class ModeFlight(Enum):

    NONE = 0x00

    READY = 0x10

    START = 0x11
    TAKEOFF = 0x12
    FLIGHT = 0x13
    LANDING = 0x14
    FLIP = 0x15
    REVERSE = 0x16

    STOP = 0x20

    ACCIDENT = 0x30
    ERROR = 0x31

    TEST = 0x40

    END_OF_TYPE = 0x41


class ModeUpdate(Enum):

    NONE = 0x00

    READY = 0x01            # 업데이트 가능 상태
    UPDATE = 0x02           # 업데이트 중
    COMPLETE = 0x03         # 업데이트 완료

    FAILED = 0x04           # 업데이트 실패(업데이트 완료까지 갔으나 body의 Crc16이 일치하지 않는 경우 등)

    NOT_AVAILABLE = 0x05    # 업데이트 불가능 상태(Debug 모드 등)
    RUN_APPLICATION = 0x06  # 어플리케이션 동작 중
    NOT_REGISTERED = 0x07   # 등록되지 않음

    END_OF_TYPE = 0x08


class ErrorFlagsForSensor(Enum):

    NONE = 0x00000000

    MOTION_NO_ANSWER = 0x00000001           # Motion 센서 응답 없음
    MOTION_WRONG_VALUE = 0x00000002         # Motion 센서 잘못된 값
    MOTION_NOT_CALIBRATED = 0x00000004      # Gyro Bias 보정이 완료되지 않음
    MOTION_CALIBRATING = 0x00000008         # Gyro Bias 보정 중

    PRESSURE_NO_ANSWER = 0x00000010         # 압력 센서 응답 없음
    PRESSURE_WRONG_VALUE = 0x00000020       # 압력 센서 잘못된 값

    RANGE_GROUND_NO_ANSWER = 0x00000100     # 바닥 거리 센서 응답 없음
    RANGE_GROUND_WRONG_VALUE = 0x00000200   # 바닥 거리 센서 잘못된 값
    RANGE_FRONT_NO_ANSWER = 0x00000400 	    # 정면 거리 센서 응답 없음
    RANGE_FRONT_WRONG_VALUE = 0x00000800    # 정면 거리 센서 잘못된 값

    FLOW_NO_ANSWER = 0x00001000             # Flow 센서 응답 없음
    FLOW_WRONG_VALUE = 0x00002000           # Flow 잘못된 값
    FLOW_CANNOT_RECOGNIZE_GROUND_IMAGE = 0x00004000     # 바닥 이미지를 인식할 수 없음

    RF_NO_ANSWER = 0x10000000               # RF 응답 없음
    RF_PAIRED = 0x20000000                  # RF 페어링 완료
    RF_CONNECTED = 0x40000000               # RF 연결됨


class ErrorFlagsForState(Enum):

    NONE = 0x00000000

    NOT_REGISTERED = 0x00000001                     # 장치 등록이 안됨
    FLASH_READ_LOCK_UNLOCKED = 0x00000002           # 플래시 메모리 읽기 Lock이 안 걸림
    BOOTLOADER_WRITE_LOCK_UNLOCKED = 0x00000004     # 부트로더 영역 쓰기 Lock이 안 걸림
    LOW_BATTERY = 0x00000008                        # Low Battery

    TAKEOFF_FAILURE_CHECK_PROPELLER_AND_MOTOR = 0x00000010  # 이륙 실패
    CHECK_PROPELLER_VIBRATION = 0x00000020          # 프로펠러 진동발생
    ATTITUDE_NOT_STABLE = 0x00000040                # 자세가 많이 기울어져 있거나 뒤집어져 있을때

    CANNOT_FLIP_LOW_BATTERY = 0x00000100            # 배터리가 30이하
    CANNOT_FLIP_TOO_HEAVY = 0x00000200              # 기체가 무거움


class FlightEvent(Enum):

    NONE = 0x00

    STOP = 0x10
    TAKEOFF = 0x11
    LANDING = 0x12

    REVERSE = 0x13

    FLIP_FRONT = 0x14
    FLIP_REAR = 0x15
    FLIP_LEFT = 0x16
    FLIP_RIGHT = 0x17

    RETURN_HOME = 0x18

    SHOT = 0x90
    UNDER_ATTACK = 0x91

    RESET_HEADING = 0xA0

    END_OF_TYPE = 0xA1


class Direction(Enum):

    NONE = 0x00

    LEFT = 0x01
    FRONT = 0x02
    RIGHT = 0x03
    REAR = 0x04

    TOP = 0x05
    BOTTOM = 0x06

    CENTER = 0x07

    END_OF_TYPE = 0x08


class Rotation(Enum):

    NONE = 0x00

    CLOCKWISE = 0x01
    COUNTERCLOCKWISE = 0x02

    END_OF_TYPE = 0x03


class SensorOrientation(Enum):

    NONE = 0x00

    NORMAL = 0x01
    REVERSE_START = 0x02
    REVERSED = 0x03

    END_OF_TYPE = 0x04


class Headless(Enum):

    NONE = 0x00

    HEADLESS = 0x01     # Headless
    NORMAL = 0x02       # Normal

    END_OF_TYPE = 0x03


class TrimDirection(Enum):

    NONE = 0x00  # 없음

    ROLL_INCREASE = 0x01    # ROLL 증가
    ROLL_DECREASE = 0x02    # ROLL 감소
    PITCH_INCREASE = 0x03   # PITCH 증가
    PITCH_DECREASE = 0x04   # PITCH 감소
    YAW_INCREASE = 0x05     # YAW 증가
    YAW_DECREASE = 0x06     # YAW 감소
    THROTTLE_INCREASE = 0x07    # THROTTLE 증가
    THROTTLE_DECREASE = 0x08    # THROTTLE 감소

    RESET = 0x09            # 전체 트림 리셋

    END_OF_TYPE = 0x0A


class ModeMovement(Enum):

    NONE = 0x00

    READY = 0x01        # Ready
    HOVERING = 0x02     # Hovering
    MOVING = 0x03       # Moving
    RETURN_HOME = 0x04  # Return Home

    END_OF_TYPE = 0x05


class CardColorIndex(Enum):

    UNKNOWN = 0x00

    WHITE = 0x01
    RED = 0x02
    YELLOW = 0x03
    GREEN = 0x04
    CYAN = 0x05
    BLUE = 0x06
    MAGENTA = 0x07
    BLACK = 0x08

    END_OF_TYPE = 0x09


class Card(Enum):

    NONE = 0x00

    WHITE_WHITE = 0x11
    WHITE_RED = 0x12
    WHITE_YELLOW = 0x13
    WHITE_GREEN = 0x14
    WHITE_CYAN = 0x15
    WHITE_BLUE = 0x16
    WHITE_MAGENTA = 0x17
    WHITE_BLACK = 0x18

    RED_WHITE = 0x21
    RED_RED = 0x22
    RED_YELLOW = 0x23
    RED_GREEN = 0x24
    RED_CYAN = 0x25
    RED_BLUE = 0x26
    RED_MAGENTA = 0x27
    RED_BLACK = 0x28

    YELLOW_WHITE = 0x31
    YELLOW_RED = 0x32
    YELLOW_YELLOW = 0x33
    YELLOW_GREEN = 0x34
    YELLOW_CYAN = 0x35
    YELLOW_BLUE = 0x36
    YELLOW_MAGENTA = 0x37
    YELLOW_BLACK = 0x38

    GREEN_WHITE = 0x41
    GREEN_RED = 0x42
    GREEN_YELLOW = 0x43
    GREEN_GREEN = 0x44
    GREEN_CYAN = 0x45
    GREEN_BLUE = 0x46
    GREEN_MAGENTA = 0x47
    GREEN_BLACK = 0x48

    CYAN_WHITE = 0x51
    CYAN_RED = 0x52
    CYAN_YELLOW = 0x53
    CYAN_GREEN = 0x54
    CYAN_CYAN = 0x55
    CYAN_BLUE = 0x56
    CYAN_MAGENTA = 0x57
    CYAN_BLACK = 0x58

    BLUE_WHITE = 0x61
    BLUE_RED = 0x62
    BLUE_YELLOW = 0x63
    BLUE_GREEN = 0x64
    BLUE_CYAN = 0x65
    BLUE_BLUE = 0x66
    BLUE_MAGENTA = 0x67
    BLUE_BLACK = 0x68

    MAGENTA_WHITE = 0x71
    MAGENTA_RED = 0x72
    MAGENTA_YELLOW = 0x73
    MAGENTA_GREEN = 0x74
    MAGENTA_CYAN = 0x75
    MAGENTA_BLUE = 0x76
    MAGENTA_MAGENTA = 0x77
    MAGENTA_BLACK = 0x78

    BLACK_WHITE = 0x81
    BLACK_RED = 0x82
    BLACK_YELLOW = 0x83
    BLACK_GREEN = 0x84
    BLACK_CYAN = 0x85
    BLACK_BLUE = 0x86
    BLACK_MAGENTA = 0x87
    BLACK_BLACK = 0x88



import serial
import binascii
import random
import queue
import threading
from threading import Thread
from time import sleep
from struct import *
import time
from serial.tools.list_ports import comports
from queue import Queue
from operator import eq
import colorama
from colorama import Fore, Back, Style

from e_drone.protocol import *
from e_drone.storage import *
from e_drone.receiver import *
from e_drone.system import *
from e_drone.crc import *


def convert_byte_array_to_string(data_array):

    if data_array is None:
        return ""

    string = ""

    if (isinstance(data_array, bytes)) or (isinstance(data_array, bytearray)) or (not isinstance(data_array, list)):
        for data in data_array:
            string += "{0:02X} ".format(data)

    return string


class Drone:

    # BaseFunctions Start
    def __init__(self, flag_check_background=True, flag_show_error_message=False, flag_show_log_message=False, flag_show_transfer_data=False, flag_show_receive_data=False):
        
        self._serialport = None
        self._buffer_queue = Queue(4096)
        self._buffer_handler = bytearray()
        self._index = 0

        self._thread = None
        self._flag_thread_run = False

        self._receiver = Receiver()

        self._flag_check_background = flag_check_background

        self._flag_show_error_message = flag_show_error_message
        self._flag_show_log_message = flag_show_log_message
        self._flag_show_transfer_data = flag_show_transfer_data
        self._flag_show_receive_data = flag_show_receive_data

        self._event_handler = EventHandler()
        self._storage_header = StorageHeader()
        self._storage = Storage()
        self._storage_count = StorageCount()
        self._parser = Parser()

        self._devices = []            # 자동 연결 시 검색된 장치 목록을 저장
        self._flag_discover = False         # 자동 연결 시 장치를 검색중인지를 표시
        self._flag_connected = False         # 자동 연결 시 장치와 연결되었는지를 알려줌

        self.time_start_program = time.time()           # 프로그램 시작 시각 기록

        self.system_time_monitor_data = 0
        self.monitor_data = []

        for i in range(0, 36):
            self.monitor_data.append(i)

        colorama.init()


    def __del__(self):
        
        self.close()


    def _receiving(self):
        while self._flag_thread_run:
            
            self._buffer_queue.put(self._serialport.read())

            # 수신 데이터 백그라운드 확인이 활성화 된 경우 데이터 자동 업데이트
            if self._flag_check_background == True:
                while self.check() != DataType.NONE:
                    pass

            #sleep(0.001)


    def is_open(self):
        if self._serialport is not None:
            return self._serialport.is_open
        else:
            return False


    def is_connected(self):
        if self.is_open() == False:
            return False
        else:
            return self._flag_connected


    def open(self, port_name="None"):
        if eq(port_name, "None"):
            nodes = comports()
            size = len(nodes)
            if size > 0:
                port_name = nodes[size - 1].device
            else:
                return False

        try:
            self._serialport = serial.Serial(port=port_name, baudrate=57600)

            if(self.is_open()):
                self._flag_thread_run = True
                self._thread = Thread(target=self._receiving, args=(), daemon=True)
                self._thread.start()

                # 로그 출력
                self._print_log("Connected.({0})".format(port_name))

                return True
            else:
                # 오류 메세지 출력
                self._print_error("Could not connect to device.")

                return False

        except:
            # 오류 메세지 출력
            self._print_error("Could not connect to device.")

            return False


    def close(self):
        # 로그 출력
        if self.is_open():
            self._print_log("Closing serial port.")

        self._print_log("Thread Flag False.")

        if self._flag_thread_run == True:
            self._flag_thread_run = False
            sleep(0.1)
        
        self._print_log("Thread Join.")

        if self._thread is not None:
            self._thread.join(timeout = 1)

        self._print_log("Port Close.")

        if self.is_open() == True:
            self._serialport.close()
            sleep(0.2)


    def make_transfer_data_array(self, header, data):
        if (header is None) or (data is None):
            return None

        if (not isinstance(header, Header)):
            return None

        if (isinstance(data, ISerializable)):
            data = data.to_array()

        crc16 = Crc16.calc(header.to_array(), 0)
        crc16 = Crc16.calc(data, crc16)

        data_array = bytearray()
        data_array.extend((0x0A, 0x55))
        data_array.extend(header.to_array())
        data_array.extend(data)
        data_array.extend(pack('H', crc16))

        return data_array


    def transfer(self, header, data):
        if not self.is_open():
            return

        data_array = self.make_transfer_data_array(header, data)

        self._serialport.write(data_array)

        # 송신 데이터 출력
        self._print_transfer_data(data_array)

        return data_array


    def check(self):
        while self._buffer_queue.empty() == False:
            data_array = self._buffer_queue.get_nowait()
            self._buffer_queue.task_done()

            if (data_array is not None) and (len(data_array) > 0):
                # 수신 데이터 출력
                self._print_receive_data(data_array)

                self._buffer_handler.extend(data_array)

        while len(self._buffer_handler) > 0:
            state_loading = self._receiver.call(self._buffer_handler.pop(0))

            # 오류 출력
            if state_loading == StateLoading.FAILURE:
                # 수신 데이터 출력(줄넘김)
                self._print_receive_data_end()

                # 오류 메세지 출력
                self._print_error(self._receiver.message)
                

            # 로그 출력
            if state_loading == StateLoading.LOADED:
                # 수신 데이터 출력(줄넘김)
                self._print_receive_data_end()

                # 로그 출력
                self._print_log(self._receiver.message)

            if self._receiver.state == StateLoading.LOADED:
                self._handler(self._receiver.header, self._receiver.data)
                return self._receiver.header.data_type

        return DataType.NONE


    def check_detail(self):
        while self._buffer_queue.empty() == False:
            data_array = self._buffer_queue.get_nowait()
            self._buffer_queue.task_done()

            if (data_array is not None) and (len(data_array) > 0):
                # 수신 데이터 출력
                self._print_receive_data(data_array)

                self._buffer_handler.extend(data_array)

        while len(self._buffer_handler) > 0:
            state_loading = self._receiver.call(self._buffer_handler.pop(0))

            # 오류 출력
            if state_loading == StateLoading.FAILURE:
                # 수신 데이터 출력(줄넘김)
                self._print_receive_data_end()

                # 오류 메세지 출력
                self._print_error(self._receiver.message)
                

            # 로그 출력
            if state_loading == StateLoading.LOADED:
                # 수신 데이터 출력(줄넘김)
                self._print_receive_data_end()

                # 로그 출력
                self._print_log(self._receiver.message)

            if self._receiver.state == StateLoading.LOADED:
                self._handler(self._receiver.header, self._receiver.data)
                return self._receiver.header, self._receiver.data

        return None, None


    def _handler(self, header, data_array):

        # 들어오는 데이터를 저장
        self._run_handler(header, data_array)

        # 콜백 이벤트 실행
        self._run_event_handler(header.data_type)

        # Monitor 데이터 처리
        self._run_handler_for_monitor(header, data_array)

        # 데이터 처리 완료 확인
        self._receiver.checked()

        return header.data_type


    def _run_handler(self, header, data_array):
        
        # 일반 데이터 처리
        if self._parser.d[header.data_type] is not None:
            self._storage_header.d[header.data_type]   = header
            self._storage.d[header.data_type]          = self._parser.d[header.data_type](data_array)
            self._storage_count.d[header.data_type]    += 1


    def _run_event_handler(self, data_type):
        if (isinstance(data_type, DataType)) and (self._event_handler.d[data_type] is not None) and (self._storage.d[data_type] is not None):
            return self._event_handler.d[data_type](self._storage.d[data_type])
        else:
            return None


    def _run_handler_for_monitor(self, header, data_array):
        
        # Monitor 데이터 처리
        # 수신 받은 데이터를 파싱하여 self.monitor_data[] 배열에 데이터를 넣음
        if header.data_type == DataType.MONITOR:
            
            monitor_header_type = MonitorHeaderType(data_array[0])

            if monitor_header_type == MonitorHeaderType.MONITOR_0:
                
                monitor0 = Monitor0.parse(data_array[1:1 + Monitor0.get_size()])

                if monitor0.monitor_dataType == monitor_dataType.F32:
                    
                    data_count = (data_array.len() - 1 - Monitor0.get_size()) / 4

                    for i in range(0, data_count):
                        
                        if monitor0.index + i < len(self.monitor_data):
                            
                            index = 1 + Monitor0.get_size() + (i * 4)
                            self.monitor_data[monitor0.index + i], = unpack('<f', data_array[index:index + 4])

            elif monitor_header_type == MonitorHeaderType.MONITOR_4:
                
                monitor4 = Monitor4.parse(data_array[1:1 + Monitor4.get_size()])

                if monitor4.monitor_dataType == monitor_dataType.F32:
                    
                    self.system_time_monitor_data = monitor4.system_time
                    
                    data_count = (data_array.len() - 1 - Monitor4.get_size()) / 4

                    for i in range(0, data_count):
                        
                        if monitor4.index + i < len(self.monitor_data):
                            
                            index = 1 + Monitor4.get_size() + (i * 4)
                            self.monitor_data[monitor4.index + i], = unpack('<f', data_array[index:index + 4])

            elif monitor_header_type == MonitorHeaderType.MONITOR_8:
                
                monitor8 = Monitor8.parse(data_array[1:1 + Monitor8.get_size()])

                if monitor8.monitor_dataType == monitor_dataType.F32:
                    
                    self.system_time_monitor_data = monitor8.system_time
                    
                    data_count = (data_array.len() - 1 - Monitor8.get_size()) / 4

                    for i in range(0, data_count):
                        
                        if monitor8.index + i < len(self.monitor_data):
                            
                            index = 1 + Monitor8.get_size() + (i * 4)
                            self.monitor_data[monitor8.index + i], = unpack('<f', data_array[index:index + 4])


    def set_event_handler(self, data_type, eventHandler):
        
        if (not isinstance(data_type, DataType)):
            return

        self._event_handler.d[data_type] = eventHandler


    def get_header(self, data_type):
    
        if (not isinstance(data_type, DataType)):
            return None

        return self._storage_header.d[data_type]


    def get_data(self, data_type):

        if (not isinstance(data_type, DataType)):
            return None

        return self._storage.d[data_type]


    def get_count(self, data_type):

        if (not isinstance(data_type, DataType)):
            return None

        return self._storage_count.d[data_type]


    def _print_log(self, message):
        
        # 로그 출력
        if self._flag_show_log_message and message is not None:
            print(Fore.GREEN + "[{0:10.03f}] {1}".format((time.time() - self.time_start_program), message) + Style.RESET_ALL)


    def _print_error(self, message):
    
        # 오류 메세지 출력
        if self._flag_show_error_message and message is not None:
            print(Fore.RED + "[{0:10.03f}] {1}".format((time.time() - self.time_start_program), message) + Style.RESET_ALL)


    def _print_transfer_data(self, data_array):
    
        # 송신 데이터 출력
        if (self._flag_show_transfer_data) and (data_array is not None) and (len(data_array) > 0):
            print(Back.YELLOW + Fore.BLACK + convert_byte_array_to_string(data_array) + Style.RESET_ALL)


    def _print_receive_data(self, data_array):
        
        # 수신 데이터 출력
        if (self._flag_show_receive_data) and (data_array is not None) and (len(data_array) > 0):
            print(Back.CYAN + Fore.BLACK + convert_byte_array_to_string(data_array) + Style.RESET_ALL, end = '')


    def _print_receive_data_end(self):
        
        # 수신 데이터 출력(줄넘김)
        if self._flag_show_receive_data:
            print("")




# BaseFunctions End



# Common Start


    def send_ping(self, device_type):
        
        if  ( not isinstance(device_type, DeviceType) ):
            return None

        header = Header()
        
        header.data_type = DataType.PING
        header.length    = Ping.get_size()
        header.from_     = DeviceType.BASE
        header.to_       = device_type

        data = Ping()

        data.system_time = 0

        return self.transfer(header, data)


    def send_request(self, device_type, data_type):
    
        if ((not isinstance(device_type, DeviceType)) or 
            (not isinstance(data_type, DataType)) ):
            return None

        header = Header()
        
        header.data_type = DataType.REQUEST
        header.length    = Request.get_size()
        header.from_     = DeviceType.BASE
        header.to_       = device_type

        data = Request()

        data.data_type   = data_type

        return self.transfer(header, data)


    def send_pairing(self, device_type, address_0, address_1, address_2, scramble, channel_0, channel_1, channel_2, channel_3):
    
        if ((not isinstance(device_type, DeviceType)) or
            (not isinstance(address_0, int)) or
            (not isinstance(address_1, int)) or
            (not isinstance(address_2, int)) or
            (not isinstance(scramble, int)) or
            (not isinstance(channel_0, int)) or
            (not isinstance(channel_1, int)) or
            (not isinstance(channel_2, int)) or
            (not isinstance(channel_3, int)) ):
            return None

        header = Header()
        
        header.data_type = DataType.PAIRING
        header.length    = Pairing.get_size()
        header.from_     = DeviceType.BASE
        header.to_       = device_type

        data = Pairing()

        data.address_0  = address_0
        data.address_1  = address_1
        data.address_2  = address_2
        data.scramble   = scramble
        data.channel_0  = channel_0
        data.channel_1  = channel_1
        data.channel_2  = channel_2
        data.channel_3  = channel_3

        return self.transfer(header, data)


# Common Start



# Control Start


    def send_takeoff(self):
        
        '''
        header = Header()
        
        header.data_type = DataType.COMMAND
        header.length    = Command.get_size()
        header.from_     = DeviceType.BASE
        header.to_       = DeviceType.DRONE

        data = Command()

        data.command_type   = CommandType.FLIGHT_EVENT
        data.option         = FlightEvent.TAKEOFF.value

        return self.transfer(header, data)
        '''

        return self.send_flight_event(FlightEvent.TAKEOFF)


    def send_landing(self):
        
        '''
        header = Header()
        
        header.data_type = DataType.COMMAND
        header.length    = Command.get_size()
        header.from_     = DeviceType.BASE
        header.to_       = DeviceType.DRONE

        data = Command()

        data.command_type   = CommandType.FLIGHT_EVENT
        data.option         = FlightEvent.LANDING.value

        return self.transfer(header, data)
        '''
        
        return self.send_flight_event(FlightEvent.LANDING)


    def send_stop(self):
        
        '''
        header = Header()
        
        header.data_type = DataType.COMMAND
        header.length   = Command.get_size()
        header.from_    = DeviceType.BASE
        header.to_      = DeviceType.DRONE

        data = Command()

        data.command_type   = CommandType.STOP
        data.option         = 0

        return self.transfer(header, data)
        '''
        
        return self.send_flight_event(FlightEvent.STOP)


    def send_control(self, roll, pitch, yaw, throttle):
        
        if ((not isinstance(roll, int)) or
            (not isinstance(pitch, int)) or
            (not isinstance(yaw, int)) or
            (not isinstance(throttle, int)) ):
            return None

        header = Header()
        
        header.data_type = DataType.CONTROL
        header.length    = ControlQuad8.get_size()
        header.from_     = DeviceType.BASE
        header.to_       = DeviceType.DRONE

        data = ControlQuad8()

        data.roll       = int(roll)
        data.pitch      = int(pitch)
        data.yaw        = int(yaw)
        data.throttle   = int(throttle)

        return self.transfer(header, data)


    def send_control_time(self, roll, pitch, yaw, throttle, time_ms):
        
        if ((not isinstance(roll, int))     or
            (not isinstance(pitch, int))    or
            (not isinstance(yaw, int))      or
            (not isinstance(throttle, int)) ):
            return None

        time_sec     = time_ms / 1000
        time_start   = time.perf_counter()

        while ((time.perf_counter() - time_start) < time_sec):
            self.send_control(roll, pitch, yaw, throttle)
            sleep(0.02)

        return self.send_control(roll, pitch, yaw, throttle)


    def send_control_position_short(self, position_x, position_y, position_z, velocity, heading, rotational_velocity):
        
        if ((not (isinstance(position_x, float) or isinstance(position_x, int))) or
            (not (isinstance(position_y, float) or isinstance(position_y, int))) or
            (not (isinstance(position_z, float) or isinstance(position_z, int))) or
            (not (isinstance(velocity, float) or isinstance(velocity, int))) or
            (not (isinstance(heading, float) or isinstance(heading, int))) or
            (not (isinstance(rotational_velocity, float) or isinstance(rotational_velocity, int)))):
            return None

        header = Header()
        
        header.data_type = DataType.CONTROL
        header.length    = ControlPositionShort.get_size()
        header.from_     = DeviceType.BASE
        header.to_       = DeviceType.DRONE

        data = ControlPositionShort()

        data.position_x             = int(position_x)
        data.position_y             = int(position_y)
        data.position_z             = int(position_z)
        data.velocity               = int(velocity)
        data.heading                = int(heading)
        data.rotational_velocity    = int(rotational_velocity)

        return self.transfer(header, data)


    def send_control_position(self, position_x, position_y, position_z, velocity, heading, rotational_velocity):
        
        if ((not (isinstance(position_x, float) or isinstance(position_x, int))) or
            (not (isinstance(position_y, float) or isinstance(position_y, int))) or
            (not (isinstance(position_z, float) or isinstance(position_z, int))) or
            (not (isinstance(velocity, float) or isinstance(velocity, int))) or
            (not (isinstance(heading, float) or isinstance(heading, int))) or
            (not (isinstance(rotational_velocity, float) or isinstance(rotational_velocity, int)))):
            return None

        header = Header()
        
        header.data_type = DataType.CONTROL
        header.length    = ControlPosition.get_size()
        header.from_     = DeviceType.BASE
        header.to_       = DeviceType.DRONE

        data = ControlPosition()

        data.position_x             = float(position_x)
        data.position_y             = float(position_y)
        data.position_z             = float(position_z)
        data.velocity               = float(velocity)
        data.heading                = int(heading)
        data.rotational_velocity    = int(rotational_velocity)

        return self.transfer(header, data)


# Control End



# Setup Start


    def send_command(self, device_type, command_type, option = 0):
        
        if ((not isinstance(device_type, DeviceType)) or
            (not isinstance(command_type, CommandType)) or
            (not isinstance(option, int)) ):
            return None

        header = Header()
        
        header.data_type = DataType.COMMAND
        header.length    = Command.get_size()
        header.from_     = DeviceType.BASE
        header.to_       = device_type

        data = Command()

        data.command_type   = command_type
        data.option         = option

        return self.transfer(header, data)


    def send_command_light_event(self, device_type, command_type, option, light_event, interval, repeat):
        
        if ((not isinstance(device_type, DeviceType)) or
            (not isinstance(command_type, CommandType)) or
            (not isinstance(option, int)) or
            (not isinstance(interval, int)) or
            (not isinstance(repeat, int))):
            return None

        header = Header()
        
        header.data_type = DataType.COMMAND
        header.length    = CommandLightEvent.get_size()
        header.from_     = DeviceType.BASE
        header.to_       = device_type

        data = CommandLightEvent()

        if ((isinstance(light_event, LightModeDrone)) or
            (isinstance(light_event, LightModeController))):
            data.event.event    = light_event.value
        elif isinstance(light_event, int):
            data.event.event    = light_event
        else:
            return None

        data.command.command_type   = command_type
        data.command.option         = option

        data.event.interval = interval
        data.event.repeat   = repeat

        return self.transfer(header, data)


    def send_command_light_event_color(self, device_type, command_type, option, light_event, interval, repeat, r, g, b):
        
        if ((not isinstance(device_type, DeviceType)) or
            (not isinstance(command_type, CommandType)) or
            (not isinstance(option, int)) or
            (not isinstance(interval, int)) or
            (not isinstance(repeat, int)) or
            (not isinstance(r, int)) or
            (not isinstance(g, int)) or
            (not isinstance(b, int))):
            return None

        header = Header()
        
        header.data_type = DataType.COMMAND
        header.length    = CommandLightEventColor.get_size()
        header.from_     = DeviceType.BASE
        header.to_       = device_type

        data = CommandLightEventColor()

        if ((isinstance(light_event, LightModeDrone)) or
            (isinstance(light_event, LightModeController))):
            data.event.event    = light_event.value
        elif isinstance(light_event, int):
            data.event.event    = light_event
        else:
            return None

        data.command.command_type   = command_type
        data.command.option         = option

        data.event.interval         = interval
        data.event.repeat           = repeat

        data.color.r                = r
        data.color.g                = g
        data.color.b                = b

        return self.transfer(header, data)


    def send_command_light_event_colors(self, device_type, command_type, option, light_event, interval, repeat, colors):
        
        if ((not isinstance(device_type, DeviceType)) or
            (not isinstance(command_type, CommandType)) or
            (not isinstance(option, int)) or
            (not isinstance(interval, int))  or
            (not isinstance(repeat, int)) or
            (not isinstance(colors, Colors))):
            return None

        header = Header()
        
        header.data_type = DataType.COMMAND
        header.length    = CommandLightEventColors.get_size()
        header.from_     = DeviceType.BASE
        header.to_       = device_type

        data = CommandLightEventColors()

        if ((isinstance(light_event, LightModeDrone)) or
            (isinstance(light_event, LightModeController))):
            data.event.event    = light_event.value
        elif isinstance(light_event, int):
            data.event.event    = light_event
        else:
            return None

        data.command.command_type   = command_type
        data.command.option         = option

        data.event.interval         = interval
        data.event.repeat           = repeat

        data.colors                 = colors

        return self.transfer(header, data)


    def send_mode_control_flight(self, mode_control_flight):
        
        if  ( not isinstance(mode_control_flight, ModeControlFlight) ):
            return None

        '''
        header = Header()
        
        header.data_type = DataType.COMMAND
        header.length    = Command.get_size()
        header.from_     = DeviceType.BASE
        header.to_       = DeviceType.DRONE

        data = Command()

        data.command_type   = CommandType.MODE_CONTROL_FLIGHT
        data.option         = mode_control_flight.value

        return self.transfer(header, data)
        '''
        return self.send_command(DeviceType.DRONE, CommandType.MODE_CONTROL_FLIGHT, mode_control_flight.value)


    def send_headless(self, headless):
        
        if  ( not isinstance(headless, Headless) ):
            return None

        '''
        header = Header()
        
        header.data_type = DataType.COMMAND
        header.length    = Command.get_size()
        header.from_     = DeviceType.BASE
        header.to_       = DeviceType.DRONE

        data = Command()

        data.command_type   = CommandType.HEADLESS
        data.option         = headless.value

        return self.transfer(header, data)
        '''
        return self.send_command(DeviceType.DRONE, CommandType.HEADLESS, headless.value)


    def send_trim(self, roll, pitch, yaw, throttle):
        
        if  ( (not isinstance(roll, int)) or (not isinstance(pitch, int)) or (not isinstance(yaw, int)) or (not isinstance(throttle, int)) ):
            return None

        header = Header()
        
        header.data_type = DataType.TRIM
        header.length    = Trim.get_size()
        header.from_     = DeviceType.BASE
        header.to_       = DeviceType.DRONE

        data = Trim()

        data.roll       = roll
        data.pitch      = pitch
        data.yaw        = yaw
        data.throttle   = throttle

        return self.transfer(header, data)


    def send_weight(self, weight):
        
        header = Header()
        
        header.data_type = DataType.WEIGHT
        header.length    = Weight.get_size()
        header.from_     = DeviceType.BASE
        header.to_       = DeviceType.DRONE

        data = Weight()

        data.weight     = weight

        return self.transfer(header, data)


    def send_lost_connection(self, time_neutral, time_landing, time_stop):
        
        header = Header()
        
        header.data_type = DataType.LOST_CONNECTION
        header.length    = LostConnection.get_size()
        header.from_     = DeviceType.BASE
        header.to_       = DeviceType.DRONE

        data = LostConnection()

        data.time_neutral    = time_neutral
        data.time_landing    = time_landing
        data.time_stop       = time_stop

        return self.transfer(header, data)


    def send_flight_event(self, flight_event):
        
        if  ( (not isinstance(flight_event, FlightEvent)) ):
            return None

        '''
        header = Header()
        
        header.data_type = DataType.COMMAND
        header.length    = Command.get_size()
        header.from_     = DeviceType.BASE
        header.to_       = DeviceType.DRONE

        data = Command()

        data.command_type   = CommandType.FLIGHT_EVENT
        data.option         = flight_event.value

        return self.transfer(header, data)
        '''
        return self.send_command(DeviceType.DRONE, CommandType.FLIGHT_EVENT, flight_event.value)


    def send_clear_bias(self):
        
        '''
        header = Header()
        
        header.data_type = DataType.COMMAND
        header.length    = Command.get_size()
        header.from_     = DeviceType.BASE
        header.to_       = DeviceType.DRONE

        data = Command()

        data.command_type   = CommandType.CLEAR_BIAS
        data.option         = 0

        return self.transfer(header, data)
        '''
        return self.send_command(DeviceType.DRONE, CommandType.CLEAR_BIAS)


    def send_clear_trim(self):
        
        '''
        header = Header()
        
        header.data_type = DataType.COMMAND
        header.length    = Command.get_size()
        header.from_     = DeviceType.BASE
        header.to_       = DeviceType.DRONE

        data = Command()

        data.command_type   = CommandType.CLEAR_TRIM
        data.option         = 0

        return self.transfer(header, data)
        '''
        return self.send_command(DeviceType.DRONE, CommandType.CLEAR_TRIM)


    def send_set_default(self, device_type):
        
        if ((not isinstance(device_type, DeviceType))):
            return None

        '''
        header = Header()
        
        header.data_type = DataType.COMMAND
        header.length    = Command.get_size()
        header.from_     = DeviceType.BASE
        header.to_       = device_type

        data = Command()

        data.command_type   = CommandType.SET_DEFAULT
        data.option         = 0

        return self.transfer(header, data)
        '''
        return self.send_command(device_type, CommandType.SET_DEFAULT)


    def send_backlight_on(self):
        
        '''
        if ((not isinstance(flagPower, bool))):
            return None

        header = Header()
        
        header.data_type = DataType.COMMAND
        header.length    = Command.get_size()
        header.from_     = DeviceType.BASE
        header.to_       = DeviceType.CONTROLLER

        data = Command()

        data.command_type   = CommandType.BACKLIGHT
        data.option         = int(flagPower)

        return self.transfer(header, data)
        '''
        return self.send_command(DeviceType.CONTROLLER, CommandType.BACKLIGHT, 1)


    def send_backlight_off(self):
        
        return self.send_command(DeviceType.CONTROLLER, CommandType.BACKLIGHT, 0)


# Setup End



# Device Start


    def send_motor(self, motor_0, motor_1, motor_2, motor_3):
        
        if ((not isinstance(motor_0, int)) or
            (not isinstance(motor_1, int)) or
            (not isinstance(motor_2, int)) or
            (not isinstance(motor_3, int))):
            return None

        header = Header()
        
        header.data_type = DataType.MOTOR
        header.length    = Motor.get_size()
        header.from_     = DeviceType.BASE
        header.to_       = DeviceType.DRONE

        data = Motor()

        data.motor[0].value     = motor_0
        data.motor[1].value     = motor_1
        data.motor[2].value     = motor_2
        data.motor[3].value     = motor_3

        return self.transfer(header, data)


    def send_motor_single(self, target, value):
        
        if ((not isinstance(target, int)) or
            (not isinstance(value, int))):
            return None

        header = Header()
        
        header.data_type = DataType.MOTOR_SINGLE
        header.length    = MotorSingle.get_size()
        header.from_     = DeviceType.BASE
        header.to_       = DeviceType.DRONE

        data = MotorSingle()

        data.target     = target
        data.value      = value

        return self.transfer(header, data)


# Device End



# Light Start


    def send_light_manual(self, device_type, flags, brightness):
        
        if ((not isinstance(device_type, DeviceType)) or
            (not isinstance(flags, int)) or
            (not isinstance(brightness, int))):
            return None

        header = Header()
        
        header.data_type = DataType.LIGHT_MANUAL
        header.length    = LightManual.get_size()
        header.from_     = DeviceType.BASE
        header.to_       = device_type

        data = LightManual()

        data.flags      = flags
        data.brightness = brightness

        return self.transfer(header, data)


    def send_light_mode(self, device_type, light_mode, interval):
        
        if ((not isinstance(device_type, DeviceType)) or
            (not isinstance(interval, int))):
            return None

        header = Header()
        
        header.data_type = DataType.LIGHT_MODE
        header.length    = LightMode.get_size()
        header.from_     = DeviceType.BASE
        header.to_       = device_type

        data = LightMode()

        if ((isinstance(light_mode, LightModeDrone)) or
            (isinstance(light_mode, LightModeController))):
            data.mode  = light_mode.value
        elif isinstance(light_mode, int):
            data.mode  = light_mode
        else:
            return None

        data.interval  = interval

        return self.transfer(header, data)


    def send_light_mode_color(self, device_type, light_mode, interval, r, g, b):
        
        if ((not isinstance(device_type, DeviceType)) or
            (not isinstance(interval, int)) or
            (not isinstance(r, int)) or
            (not isinstance(g, int)) or
            (not isinstance(b, int))):
            return None

        header = Header()
        
        header.data_type = DataType.LIGHT_MODE
        header.length    = LightModeColor.get_size()
        header.from_     = DeviceType.BASE
        header.to_       = device_type

        data = LightModeColor()

        if ((isinstance(light_mode, LightModeDrone)) or
            (isinstance(light_mode, LightModeController))):
            data.mode.mode  = light_mode.value
        elif isinstance(light_mode, int):
            data.mode.mode  = light_mode
        else:
            return None

        data.mode.interval  = interval

        data.color.r        = r
        data.color.g        = g
        data.color.b        = b

        return self.transfer(header, data)


    def send_light_mode_colors(self, device_type, light_mode, interval, colors):
        
        if ((not isinstance(device_type, DeviceType)) or
            (not isinstance(interval, int)) or
            (not isinstance(colors, Colors))):
            return None

        header = Header()
        
        header.data_type = DataType.LIGHT_MODE
        header.length    = LightModeColors.get_size()
        header.from_     = DeviceType.BASE
        header.to_       = device_type

        data = LightModeColors()

        if ((isinstance(light_mode, LightModeDrone)) or
            (isinstance(light_mode, LightModeController))):
            data.mode.mode  = light_mode.value
        elif isinstance(light_mode, int):
            data.mode.mode  = light_mode
        else:
            return None

        data.mode.interval  = interval
        data.colors         = colors

        return self.transfer(header, data)


    def send_light_event(self, device_type, light_event, interval, repeat):
        
        if ((not isinstance(device_type, DeviceType)) or
            (not isinstance(interval, int)) or
            (not isinstance(repeat, int))):
            return None

        header = Header()
        
        header.data_type = DataType.LIGHT_EVENT
        header.length    = LightEvent.get_size()
        header.from_     = DeviceType.BASE
        header.to_       = device_type

        data = LightEvent()

        if ((isinstance(light_event, LightModeDrone)) or
            (isinstance(light_event, LightModeController))):
            data.event    = light_event.value
        elif isinstance(light_event, int):
            data.event    = light_event
        else:
            return None

        data.interval = interval
        data.repeat   = repeat

        return self.transfer(header, data)


    def send_light_event_color(self, device_type, light_event, interval, repeat, r, g, b):
        
        if ((not isinstance(device_type, DeviceType)) or
            (not isinstance(interval, int)) or
            (not isinstance(repeat, int)) or
            (not isinstance(r, int)) or
            (not isinstance(g, int)) or
            (not isinstance(b, int))):
            return None

        header = Header()
        
        header.data_type = DataType.LIGHT_EVENT
        header.length    = LightEventColor.get_size()
        header.from_     = DeviceType.BASE
        header.to_       = device_type

        data = LightEventColor()

        if ((isinstance(light_event, LightModeDrone)) or
            (isinstance(light_event, LightModeController))):
            data.event.event    = light_event.value
        elif isinstance(light_event, int):
            data.event.event    = light_event
        else:
            return None

        data.event.interval = interval
        data.event.repeat   = repeat

        data.color.r        = r
        data.color.g        = g
        data.color.b        = b

        return self.transfer(header, data)


    def send_light_event_colors(self, device_type, light_event, interval, repeat, colors):
        
        if ((not isinstance(device_type, DeviceType)) or
            (not isinstance(interval, int)) or
            (not isinstance(repeat, int)) or
            (not isinstance(colors, Colors))):
            return None

        header = Header()
        
        header.data_type = DataType.LIGHT_EVENT
        header.length    = LightEventColors.get_size()
        header.from_     = DeviceType.BASE
        header.to_       = device_type

        data = LightEventColors()

        if ((isinstance(light_event, LightModeDrone)) or
            (isinstance(light_event, LightModeController))):
            data.event.event    = light_event.value
        elif isinstance(light_event, int):
            data.event.event    = light_event
        else:
            return None

        data.event.interval = interval
        data.event.repeat   = repeat

        data.colors         = colors

        return self.transfer(header, data)


    def send_light_default_color(self, device_type, light_mode, interval, r, g, b):
        
        if ((not isinstance(device_type, DeviceType)) or
            (not isinstance(interval, int)) or
            (not isinstance(r, int)) or
            (not isinstance(g, int)) or
            (not isinstance(b, int))):
            return None

        header = Header()
        
        header.data_type = DataType.LIGHT_DEFAULT
        header.length    = LightModeColor.get_size()
        header.from_     = DeviceType.BASE
        header.to_       = device_type

        data = LightModeColor()

        if ((isinstance(light_mode, LightModeDrone)) or
            (isinstance(light_mode, LightModeController))):
            data.mode.mode    = light_mode.value
        elif isinstance(light_mode, int):
            data.mode.mode    = light_mode
        else:
            return None

        data.mode.interval  = interval

        data.color.r        = r
        data.color.g        = g
        data.color.b        = b

        return self.transfer(header, data)


# Light End



# Display Start


    def send_display_clear_all(self, pixel = DisplayPixel.BLACK):
        
        if ( not isinstance(pixel, DisplayPixel) ):
            return None

        header = Header()
        
        header.data_type = DataType.DISPLAY_CLEAR
        header.length    = DisplayClearAll.get_size()
        header.from_     = DeviceType.BASE
        header.to_       = DeviceType.CONTROLLER
        
        data = DisplayClearAll()

        data.pixel      = pixel

        return self.transfer(header, data)


    def send_display_clear(self, x, y, width, height, pixel = DisplayPixel.BLACK):
        
        if ( not isinstance(pixel, DisplayPixel) ):
            return None

        header = Header()
        
        header.data_type = DataType.DISPLAY_CLEAR
        header.length    = DisplayClear.get_size()
        header.from_     = DeviceType.BASE
        header.to_       = DeviceType.CONTROLLER
        
        data = DisplayClear()

        data.x          = x
        data.y          = y
        data.width      = width
        data.height     = height
        data.pixel      = pixel

        return self.transfer(header, data)


    def send_display_invert(self, x, y, width, height):
        
        header = Header()
        
        header.data_type = DataType.DISPLAY_INVERT
        header.length    = DisplayInvert.get_size()
        header.from_     = DeviceType.BASE
        header.to_       = DeviceType.CONTROLLER
        
        data = DisplayInvert()

        data.x          = x
        data.y          = y
        data.width      = width
        data.height     = height

        return self.transfer(header, data)


    def send_display_draw_point(self, x, y, pixel = DisplayPixel.WHITE):
        
        if ( not isinstance(pixel, DisplayPixel) ):
            return None

        header = Header()
        
        header.data_type = DataType.DISPLAY_DRAW_POINT
        header.length    = DisplayDrawPoint.get_size()
        header.from_     = DeviceType.BASE
        header.to_       = DeviceType.CONTROLLER
        
        data = DisplayDrawPoint()

        data.x          = x
        data.y          = y
        data.pixel      = pixel

        return self.transfer(header, data)


    def send_display_draw_line(self, x1, y1, x2, y2, pixel = DisplayPixel.WHITE, line = DisplayLine.SOLID):
        
        if ((not isinstance(pixel, DisplayPixel)) or
            (not isinstance(line, DisplayLine)) ):
            return None

        header = Header()
        
        header.data_type = DataType.DISPLAY_DRAW_LINE
        header.length    = DisplayDrawLine.get_size()
        header.from_     = DeviceType.BASE
        header.to_       = DeviceType.CONTROLLER
        
        data = DisplayDrawLine()

        data.x1         = x1
        data.y1         = y1
        data.x2         = x2
        data.y2         = y2
        data.pixel      = pixel
        data.line       = line

        return self.transfer(header, data)


    def send_display_draw_rect(self, x, y, width, height, pixel = DisplayPixel.WHITE, flag_fill = False, line = DisplayLine.SOLID):
        
        if ((not isinstance(pixel, DisplayPixel)) or
            (not isinstance(flag_fill, bool)) or
            (not isinstance(line, DisplayLine)) ):
            return None

        header = Header()
        
        header.data_type = DataType.DISPLAY_DRAW_RECT
        header.length    = DisplayDrawRect.get_size()
        header.from_     = DeviceType.BASE
        header.to_       = DeviceType.CONTROLLER
        
        data = DisplayDrawRect()

        data.x          = x
        data.y          = y
        data.width      = width
        data.height     = height
        data.pixel      = pixel
        data.flag_fill  = flag_fill
        data.line       = line

        return self.transfer(header, data)


    def send_display_draw_circle(self, x, y, radius, pixel = DisplayPixel.WHITE, flag_fill = True):
        
        if ((not isinstance(pixel, DisplayPixel)) or
            (not isinstance(flag_fill, bool))):
            return None

        header = Header()
        
        header.data_type = DataType.DISPLAY_DRAW_CIRCLE
        header.length    = DisplayDrawCircle.get_size()
        header.from_     = DeviceType.BASE
        header.to_       = DeviceType.CONTROLLER
        
        data = DisplayDrawCircle()

        data.x          = x
        data.y          = y
        data.radius     = radius
        data.pixel      = pixel
        data.flag_fill  = flag_fill

        return self.transfer(header, data)


    def send_display_draw_string(self, x, y, message, font = DisplayFont.LIBERATION_MONO_5X8, pixel = DisplayPixel.WHITE):
        
        if ((not isinstance(font, DisplayFont)) or
            (not isinstance(pixel, DisplayPixel)) or
            (not isinstance(message, str))):
            return None

        header = Header()
        
        header.data_type = DataType.DISPLAY_DRAW_STRING
        header.length    = DisplayDrawString.get_size() + len(message)
        header.from_     = DeviceType.BASE
        header.to_       = DeviceType.CONTROLLER
        
        data = DisplayDrawString()

        data.x          = x
        data.y          = y
        data.font       = font
        data.pixel      = pixel
        data.message    = message

        return self.transfer(header, data)


    def send_display_draw_string_align(self, x_start, x_end, y, message, align = DisplayAlign.CENTER, font = DisplayFont.LIBERATION_MONO_5X8, pixel = DisplayPixel.WHITE):
        
        if ((not isinstance(align, DisplayAlign)) or
            (not isinstance(font, DisplayFont)) or
            (not isinstance(pixel, DisplayPixel)) or
            (not isinstance(message, str))):
            return None

        header = Header()
        
        header.data_type = DataType.DISPLAY_DRAW_STRING_ALIGN
        header.length    = DisplayDrawStringAlign.get_size() + len(message)
        header.from_     = DeviceType.BASE
        header.to_       = DeviceType.CONTROLLER
        
        data = DisplayDrawStringAlign()

        data.x_start    = x_start
        data.x_end      = x_end
        data.y          = y
        data.align      = align
        data.font       = font
        data.pixel      = pixel
        data.message    = message

        return self.transfer(header, data)


# Display End



# Buzzer Start


    def send_buzzer(self, device_type, mode, value, time):
        
        if ((not isinstance(device_type, DeviceType)) or
            (not isinstance(mode, BuzzerMode)) or
            (not isinstance(value, int)) or
            (not isinstance(time, int))):
            return None

        header = Header()
        
        header.data_type = DataType.BUZZER
        header.length    = Buzzer.get_size()
        header.from_     = DeviceType.BASE
        header.to_       = device_type
        
        data = Buzzer()

        data.mode       = mode
        data.value      = value
        data.time       = time

        return self.transfer(header, data)


    def send_buzzer_mute(self, device_type, time):
        
        if ((not isinstance(device_type, DeviceType)) or
            (not isinstance(time, int))):
            return None

        header = Header()
        
        header.data_type = DataType.BUZZER
        header.length    = Buzzer.get_size()
        header.from_     = DeviceType.BASE
        header.to_       = device_type
        
        data = Buzzer()

        data.mode       = BuzzerMode.MUTE
        data.value      = BuzzerScale.MUTE.value
        data.time       = time

        return self.transfer(header, data)


    def send_buzzer_mute_reserve(self, device_type, time):
        
        if ((not isinstance(device_type, DeviceType)) or
            (not isinstance(time, int))):
            return None

        header = Header()
        
        header.data_type = DataType.BUZZER
        header.length    = Buzzer.get_size()
        header.from_     = DeviceType.BASE
        header.to_       = device_type
        
        data = Buzzer()

        data.mode       = BuzzerMode.MUTE_RESERVE
        data.value      = BuzzerScale.MUTE.value
        data.time       = time

        return self.transfer(header, data)


    def send_buzzer_scale(self, device_type, scale, time):
        
        if ((not isinstance(device_type, DeviceType)) or
            (not isinstance(scale, BuzzerScale)) or
            (not isinstance(time, int))):
            return None

        header = Header()
        
        header.data_type = DataType.BUZZER
        header.length    = Buzzer.get_size()
        header.from_     = DeviceType.BASE
        header.to_       = device_type
        
        data = Buzzer()

        data.mode       = BuzzerMode.SCALE
        data.value      = scale.value
        data.time       = time

        return self.transfer(header, data)


    def send_buzzer_scale_reserve(self, device_type, scale, time):
        
        if ((not isinstance(device_type, DeviceType)) or
            (not isinstance(scale, BuzzerScale)) or
            (not isinstance(time, int))):
            return None

        header = Header()
        
        header.data_type = DataType.BUZZER
        header.length    = Buzzer.get_size()
        header.from_     = DeviceType.BASE
        header.to_       = device_type
        
        data = Buzzer()

        data.mode       = BuzzerMode.SCALE_RESERVE
        data.value      = scale.value
        data.time       = time

        return self.transfer(header, data)


    def send_buzzer_hz(self, device_type, hz, time):
        
        if ((not isinstance(device_type, DeviceType)) or
            (not isinstance(hz, int)) or
            (not isinstance(time, int))):
            return None

        header = Header()
        
        header.data_type = DataType.BUZZER
        header.length    = Buzzer.get_size()
        header.from_     = DeviceType.BASE
        header.to_       = device_type
        
        data = Buzzer()

        data.mode       = BuzzerMode.HZ
        data.value      = hz
        data.time       = time

        return self.transfer(header, data)


    def send_buzzer_hz_reserve(self, device_type, hz, time):
        
        if ((not isinstance(device_type, DeviceType)) or
            (not isinstance(hz, int)) or
            (not isinstance(time, int))):
            return None

        header = Header()
        
        header.data_type = DataType.BUZZER
        header.length    = Buzzer.get_size()
        header.from_     = DeviceType.BASE
        header.to_       = device_type
        
        data = Buzzer()

        data.mode       = BuzzerMode.HZ_RESERVE
        data.value      = hz
        data.time       = time

        return self.transfer(header, data)


# Buzzer End



# Vibrator Start


    def send_vibrator(self, on, off, total):
        
        if ((not isinstance(on, int)) or
            (not isinstance(off, int)) or
            (not isinstance(total, int))):
            return None

        header = Header()
        
        header.data_type = DataType.VIBRATOR
        header.length    = Vibrator.get_size()
        header.from_     = DeviceType.BASE
        header.to_       = DeviceType.CONTROLLER
        
        data = Vibrator()

        data.mode       = VibratorMode.INSTANTLY
        data.on         = on
        data.off        = off
        data.total      = total

        return self.transfer(header, data)


    def send_vibrator_reserve(self, on, off, total):
        
        if ((not isinstance(on, int)) or
            (not isinstance(off, int)) or
            (not isinstance(total, int))):
            return None

        header = Header()
        
        header.data_type = DataType.VIBRATOR
        header.length    = Vibrator.get_size()
        header.from_     = DeviceType.BASE
        header.to_       = DeviceType.CONTROLLER
        
        data = Vibrator()

        data.mode       = VibratorMode.CONTINUALLY
        data.on         = on
        data.off        = off
        data.total      = total

        return self.transfer(header, data)


# Vibrator End



# Update Start




# Update End



import time

from e_drone.protocol import *
from e_drone.crc import Crc16



# 데이터 수신 상태
class StateLoading(Enum):
    
    READY           = 0x00      # 수신 대기
    RECEIVING       = 0x01      # 수신중
    LOADED          = 0x02      # 수신 완료 후 명령어 보관소에 대기중
    FAILURE         = 0x03      # 수신 실패



# 데이터 섹션 구분
class Section(Enum):

    START           = 0x00      # 전송 시작 코드
    HEADER          = 0x01      # 헤더
    DATA            = 0x02      # 데이터
    END             = 0x03      # 데이터 확인



class Receiver:


    def __init__(self):
        
        self.state                    = StateLoading.READY
        self.section_old              = Section.END
        self.section                  = Section.START
        self.index                    = 0

        self.header                   = Header()
        self.time_receive_start       = 0
        self.time_receive_complete    = 0

        self._buffer                  = bytearray()
        self.data                     = bytearray()

        self.crc16_received           = 0
        self.crc16_calculated         = 0

        self.message                  = None



    def call(self, data):
        
        now = time.perf_counter() * 1000

        self.message = None


        # First Step
        if self.state == StateLoading.FAILURE:
            self.state = StateLoading.READY


        # Second Step
        if self.state == StateLoading.READY:
            self.section = Section.START
            self.index = 0

        elif self.state == StateLoading.RECEIVING:
            # 데이터 수신을 시작한지 600ms 시간이 지난 경우 오류 출력
            if (self.time_receive_start + 600) < now:
                self.state = StateLoading.FAILURE
                self.message = "Error / Receiver / StateLoading.RECEIVING / Time over."
                return self.state

        elif self.state == StateLoading.LOADED:
            return self.state


        # Third Step
        if self.section != self.section_old:
            self.index = 0
            self.section_old = self.section
        

        # Third Step
        if self.section == Section.START:
            if self.index == 0:
                if data == 0x0A:
                    self.state = StateLoading.RECEIVING

                else:
                    self.state = StateLoading.FAILURE
                    return self.state
                self.time_receive_start = now

            elif self.index == 1:
                if data != 0x55:
                    self.state = StateLoading.FAILURE
                    return self.state
                else:
                    self.section = Section.HEADER
            else:
                self.state = StateLoading.FAILURE
                self.message = "Error / Receiver / Section.START / Index over."
                return self.state
        
        elif self.section == Section.HEADER:
            
            if self.index == 0:
                self.header = Header()
                
                try:
                    self.header.data_type = DataType(data)

                except:
                    self.state = StateLoading.FAILURE
                    self.message = "Error / Receiver / Section.HEADER / DataType Error. 0x{0:02X}".format(data)
                    return self.state

                self.crc16_calculated = Crc16.calc(data, 0)

            elif self.index == 1:
                self.header.length = data
                self.crc16_calculated = Crc16.calc(data, self.crc16_calculated)

                if self.header.length > 128:
                    self.state = StateLoading.FAILURE
                    self.message = "Error / Receiver / Section.HEADER / Data length is longer than 128. [{0}]".format(self.header.length)
                    return self.state

            elif self.index == 2:
                try:
                    self.header.from_ = DeviceType(data)

                except:
                    self.state = StateLoading.FAILURE
                    self.message = "Error / Receiver / Section.HEADER / DeviceType Error. 0x{0:02X}".format(data)
                    return self.state

                self.crc16_calculated = Crc16.calc(data, self.crc16_calculated)

            elif self.index == 3:
                try:
                    self.header.to_ = DeviceType(data)

                except:
                    self.state = StateLoading.FAILURE
                    self.message = "Error / Receiver / Section.HEADER / DeviceType Error. 0x{0:02X}".format(data)
                    return self.state

                self.crc16_calculated = Crc16.calc(data, self.crc16_calculated)

                if self.header.length == 0:
                    self.section = Section.END

                else:
                    self.section = Section.DATA
                    self._buffer.clear()

            else:
                self.state = StateLoading.FAILURE
                self.message = "Error / Receiver / Section.HEADER / Index over."
                return self.state
        
        elif self.section == Section.DATA:
            self._buffer.append(data)
            self.crc16_calculated = Crc16.calc(data, self.crc16_calculated)

            if (self.index == self.header.length - 1):
                self.section = Section.END
        
        elif self.section == Section.END:
            if   self.index == 0:
                self.crc16_received = data

            elif self.index == 1:
                self.crc16_received = (data << 8) | self.crc16_received

                if self.crc16_received == self.crc16_calculated:
                    self.data = self._buffer.copy()
                    self.time_receive_complete = now
                    self.state = StateLoading.LOADED
                    self.message = "Success / Receiver / Section.END / Receive complete / {0} / [receive: 0x{1:04X}]".format(self.header.data_type, self.crc16_received)
                    return self.state

                else:
                    self.state = StateLoading.FAILURE
                    self.message = "Error / Receiver / Section.END / CRC Error / {0} / [receive: 0x{1:04X}, calculate: 0x{2:04X}]".format(self.header.data_type, self.crc16_received, self.crc16_calculated)
                    return self.state

            else:
                self.state = StateLoading.FAILURE
                self.message = "Error / Receiver / Section.END / Index over."
                return self.state

        else:
            self.state = StateLoading.FAILURE
            self.message = "Error / Receiver / Section over."
            return self.state


        #Forth Step
        if self.state == StateLoading.RECEIVING:
            self.index += 1

        return self.state



    def checked(self):
        self.state = StateLoading.READY




import os
import abc
import sys
from struct import *
from enum import Enum
from urllib.request import urlopen
from time import sleep

from e_drone.drone import *


# Firmware Start


class FirmwareHeader():

    def __init__(self):
        self.model_number       = 0
        self.version            = 0
        self.length             = 0

        self.year               = 0
        self.month              = 0
        self.day                = 0

        self.version_major      = 0
        self.version_minor      = 0
        self.version_build      = 0


    @classmethod
    def get_size(cls):
        return 16


    @classmethod
    def parse(cls, data_array):
        data = FirmwareHeader()
        
        if len(data_array) != cls.get_size():
            return None
        
        data.model_number, data.version, data.length, data.year, data.month, data.day = unpack('<IIIHBB', data_array)

        data.model_number = ModelNumber(data.model_number)

        data.version_major = (data.version >> 24) & 0xFF
        data.version_minor = (data.version >> 16) & 0xFF
        data.version_build = (data.version & 0xFFFF)

        return data



class Firmware():

    def __init__(self, url = None):

        if url is None:
            self.url            = 0         # 펌웨어 URL
            self.resource       = None      # 펌웨어 전체 파일
            self.header         = None      # 펌웨어 헤더
            self.length         = 0         # 펌웨어 전체 파일의 길이

            self.raw_header      = None      # 헤더 배열
            self.string_header   = None      # 헤더 배열을 HEX 문자열로 변환한 것
        else:
            self.open(url)


    def open(self, url):

        self.url = url
        
        with urlopen(self.url) as res:
            self.resource   = res.read()
            self.length     = len(self.resource)
            self.raw_header = self.resource[0:16]
            self.header     = FirmwareHeader.parse(self.raw_header)

            self.string_header = ""
            for data in self.raw_header:
                self.string_header += "{0:02X} ".format(data)

            print(Fore.CYAN + "  - {0}".format(self.header.model_number) + Style.RESET_ALL)
            print("    Header Hex : {0}".format(self.string_header))
            print("     File Size : {0} bytes".format(self.length))
            print("       Version : {0}.{1}.{2}".format(self.header.version_major, self.header.version_minor, self.header.version_build))
            print("          Date : {0}.{1}.{2}".format(self.header.year, self.header.month, self.header.day))
            print("        Length : {0}\n".format(self.header.length))


# Firmware End




# Updater Start

class Updater:


    def __init__(self):

        self.model_number         = None
        self.device_type          = None
        self.mode_update          = ModeUpdate.NONE
        self.index_block_next     = 0
        self.flag_updated         = False
        self.flag_update_complete = False



    def eventInformation(self, information):

        self.mode_update         = information.mode_update
        self.flag_updated        = True
        self.model_number        = information.model_number
        self.device_type         = DeviceType(((self.model_number.value >> 8) & 0xFF))
        
        if information.mode_update == ModeUpdate.COMPLETE:
            self.flag_update_complete = True
        else:
            print(Fore.YELLOW + "* Connected Device : {0}".format(self.device_type) + Style.RESET_ALL)
            print("  Model Number : {0}".format(information.model_number))
            print("       Version : {0}.{1}.{2} ({3} / 0x{3:08X})".format(information.version.major, information.version.minor, information.version.build, information.version.v))
            print("  Release Date : {0}.{1}.{2}".format(information.year, information.month, information.day))
            print("   Mode Update : {0}\n".format(information.mode_update))



    def eventUpdateLocation(self, update_location):

        self.flag_updated        = True
        self.index_block_next    = update_location.index_block_next
        #print("* eventUpdateLocation({0})\n".format(update_location.index_block_next))



    def update(self):

        colorama.init()

        print(Back.WHITE + Fore.BLUE + " FIRMWARE UPGRADE " + Style.RESET_ALL)
        print("")

        print(Fore.YELLOW + "* Firmware loading." + Style.RESET_ALL)
        print("")
        
        firmware = []
        firmware.append(Firmware("https://s3.ap-northeast-2.amazonaws.com/byrobot/fw_drone_4_drone_p5_latest.eb"))
        firmware.append(Firmware("https://s3.ap-northeast-2.amazonaws.com/byrobot/fw_drone_4_controller_p2_latest.eb"))

        drone = Drone()
        if drone.open() == False:
            print(Fore.RED + "* Error : Unable to open serial port." + Style.RESET_ALL)
            sys.exit(1)


        # 이벤트 핸들링 함수 등록
        drone.set_event_handler(DataType.INFORMATION, self.eventInformation)
        drone.set_event_handler(DataType.UPDATE_LOCATION, self.eventUpdateLocation)


        # 변수
        flag_run             = True
        count_error          = 0
        time_transfer_next   = 0
        time_draw_next       = 0     # 업데이트 상태 다음 갱신 시각


        # 연결된 장치 확인
        drone.send_request(DeviceType.DRONE, DataType.INFORMATION)
        sleep(0.2)

        drone.send_request(DeviceType.CONTROLLER, DataType.INFORMATION)
        sleep(0.2)


        if self.device_type is None:
            print(Fore.RED + "* Error : No Answer." + Style.RESET_ALL)
            sys.exit(1)


        # 헤더 만들기
        header = Header()
        header.data_type = DataType.UPDATE
        header.length    = 18
        header.from_     = DeviceType.UPDATER
        header.to_       = self.device_type


        # 업데이트 위치 요청
        drone.send_request(header.to_, DataType.UPDATE_LOCATION)
        sleep(0.1)

        drone.send_request(header.to_, DataType.UPDATE_LOCATION)
        sleep(0.1)


        # 펌웨어 업데이트
        for fw in firmware:

            if self.model_number == fw.header.model_number:

                # 펌웨어의 모델 번호와 일치하는 드론이 있는 경우

                if self.mode_update == ModeUpdate.READY or self.mode_update == ModeUpdate.UPDATE:

                    while flag_run:

                        sleep(0.001)
                        now = time.perf_counter() * 1000

                        if (self.flag_updated == True) or (time_transfer_next < now):

                            if self.index_block_next == 0:
                                time_transfer_next = now + 2400
                            else:
                                time_transfer_next = now + 100

                            # 에러 카운트
                            if self.flag_updated == False:

                                count_error = count_error + 1

                                # 오류가 과도하게 누적된 경우 업데이트 취소
                                if count_error > 30:
                                    print(Fore.RED + "* Error : No response." + Style.RESET_ALL)
                                    flag_run = False
                            
                            else:
                                count_error = 0


                            index = self.index_block_next * 16

                            # 업데이트 할 위치를 넘어서는 경우 종료
                            if index + 16 > fw.length:
                                print(Fore.RED + "* Error : Index Over." + Style.RESET_ALL)
                                flag_run = False
                                break

                            # 업데이트가 완료된 경우 종료
                            if self.flag_update_complete == True:
                                sleep(1)
                                print("\n\n" + Fore.GREEN + "  Update Complete." + Style.RESET_ALL)
                                flag_run = False
                                break

                            data = bytearray(2)
                            data[0] = self.index_block_next & 0xFF
                            data[1] = (self.index_block_next >> 8) & 0xFF
                            data.extend(fw.resource[index : index + 16])
                            drone.transfer(header, data)
                            
                            self.flag_updated = False

                            # 진행률 표시
                            if (time_draw_next < now) or (fw.length - index < 128):

                                time_draw_next    = now + 73
                                percentage      = index * 100 / fw.length
                                print(Fore.YELLOW + "\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b{0:8.1f}%".format(percentage) + Style.RESET_ALL, end='')
                else:
                    print(Fore.RED + "* Error : Firmware update is not available." + Style.RESET_ALL)

                break

        drone.close()


# Updater End




# Main Start


if __name__ == '__main__':

    updater = Updater()

    updater.update()


# Main End




import time

from e_drone.protocol import *
from e_drone.system import *


# EventHandler
class EventHandler:

    def __init__(self):
        self.d = dict.fromkeys(list(DataType))



# StorageHeader
class StorageHeader:

    def __init__(self):
        self.d = dict.fromkeys(list(DataType))



# Storage
class Storage:

    def __init__(self):
        self.d = dict.fromkeys(list(DataType))



# Storage Count
class StorageCount:

    def __init__(self):
        self.d = dict.fromkeys(list(DataType))
        
        for key in self.d:
            self.d[key] = 0



# Storage
class Parser:

    def __init__(self):
        self.d = dict.fromkeys(list(DataType))

        self.d[DataType.PING]               = Ping.parse
        self.d[DataType.ACK]                = Ack.parse
        self.d[DataType.ERROR]              = Error.parse
        self.d[DataType.MESSAGE]            = Message.parse
        self.d[DataType.ADDRESS]            = Address.parse
        self.d[DataType.INFORMATION]        = Information.parse
        self.d[DataType.UPDATE_LOCATION]    = UpdateLocation.parse
        self.d[DataType.SYSTEM_INFORMATION] = SystemInformation.parse
        self.d[DataType.REGISTRATION]       = RegistrationInformation.parse

        self.d[DataType.PAIRING]            = Pairing.parse
        self.d[DataType.RSSI]               = Rssi.parse

        self.d[DataType.RAW_MOTION]         = RawMotion.parse
        self.d[DataType.RAW_FLOW]           = RawFlow.parse

        self.d[DataType.STATE]              = State.parse
        self.d[DataType.ATTITUDE]           = Attitude.parse
        self.d[DataType.POSITION]           = Position.parse
        self.d[DataType.ALTITUDE]           = Altitude.parse
        self.d[DataType.MOTION]             = Motion.parse
        self.d[DataType.RANGE]              = Range.parse
        self.d[DataType.FLOW]               = Flow.parse

        self.d[DataType.COUNT]              = Count.parse
        self.d[DataType.BIAS]               = Bias.parse
        self.d[DataType.TRIM]               = Trim.parse
        self.d[DataType.WEIGHT]             = Weight.parse

        self.d[DataType.BUTTON]             = Button.parse
        self.d[DataType.JOYSTICK]           = Joystick.parse

        self.d[DataType.CARD_CLASSIFY]       = CardClassify.parse
        self.d[DataType.CARD_RANGE]          = CardRange.parse
        self.d[DataType.CARD_RAW]            = CardRaw.parse
        self.d[DataType.CARD_COLOR]          = CardColor.parse
        self.d[DataType.CARD_LIST]           = CardList.parse
        self.d[DataType.CARD_FUNCTION_LIST]  = CardList.parse

        self.d[DataType.INFORMATION_ASSEMBLED_FOR_CONTROLLER]  = InformationAssembledForController.parse
        self.d[DataType.INFORMATION_ASSEMBLED_FOR_ENTRY]       = InformationAssembledForEntry.parse


